package org.dasein.cloud.jclouds.terremark;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.services.address.AddressServices;
import org.dasein.cloud.services.address.AddressType;
import org.dasein.cloud.services.address.IpAddress;
import org.dasein.cloud.services.address.IpForwardingRule;
import org.dasein.cloud.services.dc.DataCenter;
import org.dasein.cloud.services.firewall.Protocol;
import org.dasein.cloud.services.server.Server;
import org.jclouds.rest.RestContext;
import org.jclouds.vcloud.terremark.TerremarkVCloudExpressAsyncClient;
import org.jclouds.vcloud.terremark.TerremarkVCloudExpressClient;
import org.jclouds.vcloud.terremark.domain.InternetService;
import org.jclouds.vcloud.terremark.domain.Node;
import org.jclouds.vcloud.terremark.domain.PublicIpAddress;
import org.jclouds.vcloud.terremark.options.AddInternetServiceOptions;
import org.jclouds.vcloud.terremark.options.AddNodeOptions;

public class PublicIp implements AddressServices {
    static private final Logger logger = Logger.getLogger(PublicIp.class);
    
    private Terremark provider = null;
    
    PublicIp(Terremark provider) {
        this.provider = provider;
    }
    
    @Override
    public void assign(String addressId, String toServerId) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }

    void clear(String serverId) throws InternalException, CloudException {
        for( IpAddress ip : listPublicIpPool(false) ) {
            for( IpForwardingRule rule : listRules(ip.getProviderIpAddressId()) ) {
                if( rule.getServerId() != null && rule.getServerId().equals(serverId) ) {
                    stopForward(rule.getProviderRuleId());
                }
            }
        }
    }

    @Override
    public String forward(String addressId, int publicPort, Protocol protocol, int privatePort, String onServerId) throws InternalException, CloudException {
        org.jclouds.vcloud.terremark.domain.Protocol proto;
        String privateAddress = null;
        Server server;
        
        server = provider.getServerServices().getServer(onServerId);
        for( String address : server.getPrivateIpAddresses() ) {
            privateAddress = address;
            break;
        }
        if( privateAddress == null ) {
            logger.error("No private address exists for " + server.getProviderServerId() + "/" + server.getName());
            throw new CloudException("No private address exists for " + server.getProviderServerId() + "/" + server.getName());
        }
        switch( protocol ) {
            case TCP: proto = org.jclouds.vcloud.terremark.domain.Protocol.TCP; break;
            case UDP: proto = org.jclouds.vcloud.terremark.domain.Protocol.UDP; break;
            default: throw new CloudException("Terremark does not support ICMP forwarding.");
        }
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            Collection<InternetService> services;
            
            try {
                services = ctx.getApi().getInternetServicesOnPublicIp(URI.create(addressId));
            }
            catch( RuntimeException e ) {
                logger.error("Could not load services on IP " + addressId);
                e.printStackTrace();
                throw new CloudException(e);
            }
            
            InternetService svc = null;
            Node node = null;
    
            for( InternetService service : services ) {
                if( service.getPort() == publicPort ) {
                    svc = service;
                    break;
                }
            }
            if( svc == null ) {
                try { 
                    svc = ctx.getApi().addInternetServiceToExistingIp(URI.create(addressId), "CUSTOM", proto, publicPort, new AddInternetServiceOptions[0]);
                }
                catch( RuntimeException e ) {
                    logger.error("Could not create internet service for IP address " + addressId + ": " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
            }
            else {
                try {
                    for( Node n : ctx.getApi().getNodes(svc.getId()) ) {
                        if( n.getPort() == privatePort ) {
                            node = n;
                            break;
                        }
                    }
                }
                catch( RuntimeException e ) {
                    logger.warn("Could not identify any existing node: " + e.getMessage());
                }
            }
            if( node != null ) {
                try {
                    ctx.getApi().deleteNode(node.getId());
                }
                catch( RuntimeException e ) {
                    logger.warn("Unable to delete node " + node.getId() + ": " + e.getMessage());
                }
            }
            try {
                node = ctx.getApi().addNode(svc.getId(), privateAddress, "Forward", privatePort, new AddNodeOptions[0]);
            }
            catch( RuntimeException e ) {
                logger.error("Failed to add node to internet service: " + e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
            return String.valueOf(node.getId());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public IpAddress getIpAddress(String addressId) throws InternalException, CloudException {
        for( IpAddress address : listPublicIpPool(false) ) {
            if( address.getProviderIpAddressId().equals(addressId) ) {
                return address;
            }
        }
        return null;
    }

    @Override
    public String getProviderTermForIpAddress(Locale locale) {
        return "IP address";
    }
    
    private String getServerAt(String address) throws InternalException, CloudException {
        for( Server server : provider.getServerServices().list() ) {
            for( String privateAddress : server.getPrivateIpAddresses() ) {
                if( privateAddress.equals(address) ) {
                    return server.getProviderServerId();
                }
            }
        }
        return null;
    }

    @Override
    public boolean isAssigned(AddressType type) {
        return false;
    }

    @Override
    public boolean isForwarding() {
        return true;
    }

    @Override
    public boolean isRequestable(AddressType type) {
        return type.equals(AddressType.PUBLIC);
    }
    
    public boolean isSubscribed() {
        return true;
    }
    
    @Override
    public Collection<IpAddress> listPrivateIpPool(boolean unassignedOnly) throws InternalException, CloudException {
        return Collections.emptyList();
    }
    
    @Override
    public Collection<IpAddress> listPublicIpPool(boolean unassignedOnly) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        
        try {
            ArrayList<IpAddress> matches = new ArrayList<IpAddress>();
            Collection<PublicIpAddress> addresses;
            
            try {
                addresses = ctx.getApi().getPublicIpsAssociatedWithVDC(ctx.getApi().findVDCInOrgNamed(null, null).getHref());
            }
            catch( RuntimeException e ) {
                logger.error(e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
            for( PublicIpAddress addr : addresses ) {
                IpAddress ip = new IpAddress();
                
                ip.setAddress(addr.getAddress());
                ip.setIpAddressId(String.valueOf(addr.getId()));
                ip.setRegionId(provider.getContext().getRegionId());
                ip.setServerId(null);
                ip.setProviderLoadBalancerId(null);
                ip.setAddressType(AddressType.PUBLIC);
                matches.add(ip);
            }
            return matches;
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public Collection<IpForwardingRule> listRules(String addressId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        
        try {
            ArrayList<IpForwardingRule> rules = new ArrayList<IpForwardingRule>();
            Collection<InternetService> services;
            
            try {
                services = ctx.getApi().getInternetServicesOnPublicIp(URI.create(addressId));
            }
            catch( RuntimeException e ) {
                logger.error("Could not load services on IP " + addressId);
                e.printStackTrace();
                throw new CloudException(e);
            }
            for( InternetService svc : services ) {
                for( Node node : ctx.getApi().getNodes(svc.getId()) ) {
                    IpForwardingRule rule = new IpForwardingRule();
                    
                    rule.setAddressId(addressId);
                    rule.setPrivatePort(node.getPort());
                    rule.setProtocol(toDaseinProtocol(svc.getProtocol()));
                    rule.setProviderRuleId(String.valueOf(node.getId()));
                    rule.setPublicPort(svc.getPort());
                    rule.setServerId(getServerAt(node.getIpAddress()));
                    rules.add(rule);
                }
            }
            return rules;
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void releaseFromPool(String addressId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            Collection<InternetService> services;
            
            try {
                services = ctx.getApi().getInternetServicesOnPublicIp(URI.create(addressId));
            }
            catch( RuntimeException e ) {
                logger.error("Could not load services on IP " + addressId);
                e.printStackTrace();
                throw new CloudException(e);
            }
            for( InternetService svc : services ) {
                for( Node node : ctx.getApi().getNodes(svc.getId()) ) {
                    if( node != null ) {
                        logger.warn("Cannot delete an IP address with existing forwarding rules.");
                        throw new CloudException("Cannot delete an IP address with existing forwarding rules.");
                    }
                }
                ctx.getApi().deleteInternetService(svc.getId());
            }
            ctx.getApi().deletePublicIp(URI.create(addressId));
        }
        finally {
            ctx.close();
        }
    }
    
    @Override
    public void releaseFromServer(String addressId) throws InternalException, CloudException {
        throw new OperationNotSupportedException();        
    }

    @Override
    public String request(AddressType type) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        
        try {
            InternetService svc = null;
            
            try {
                for( DataCenter dc : provider.getDataCenterServices().listDataCenters(provider.getContext().getRegionId()) ) {
                    svc = ctx.getApi().addInternetServiceToVDC(URI.create(dc.getProviderDataCenterId()), "SSH", org.jclouds.vcloud.terremark.domain.Protocol.TCP, 22);
                    break;
                }
            }
            catch( RuntimeException e ) {
                throw new CloudException(e);
            }
            return (svc == null ? null : String.valueOf(svc.getPublicIpAddress().getId()));
        }
        finally {
            ctx.close();
        }
    }
    
    @Override
    public void stopForward(String ruleId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            ctx.getApi().deleteNode(URI.create(ruleId));
        }
        finally {
            ctx.close();
        }
    }
    
    private Protocol toDaseinProtocol(org.jclouds.vcloud.terremark.domain.Protocol jcloudsProtocol) {
        switch( jcloudsProtocol ) {
            case TCP: case HTTP: case HTTPS: case FTP: return Protocol.TCP;
            case UDP: return Protocol.UDP;
            default: return Protocol.TCP;
        }
    }
}
