package org.dasein.cloud.jclouds.terremark;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Locale;
import java.util.Random;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.services.dc.DataCenter;
import org.dasein.cloud.services.dc.Region;
import org.dasein.cloud.services.image.ServerImage;
import org.dasein.cloud.services.server.Architecture;
import org.dasein.cloud.services.server.Platform;
import org.dasein.cloud.services.server.Server;
import org.dasein.cloud.services.server.ServerPersistence;
import org.dasein.cloud.services.server.ServerServices;
import org.dasein.cloud.services.server.ServerSize;
import org.dasein.cloud.services.server.ServerState;
import org.dasein.cloud.services.server.ServerStatistics;
import org.jclouds.rest.RestContext;
import org.jclouds.vcloud.VCloudMediaType;
import org.jclouds.vcloud.domain.ReferenceType;
import org.jclouds.vcloud.domain.Task;
import org.jclouds.vcloud.domain.TaskStatus;
import org.jclouds.vcloud.domain.VCloudExpressVApp;
import org.jclouds.vcloud.domain.VDC;
import org.jclouds.vcloud.domain.ovf.ResourceAllocation;
import org.jclouds.vcloud.domain.ovf.ResourceType;
import org.jclouds.vcloud.options.CloneVAppOptions;
import org.jclouds.vcloud.terremark.TerremarkVCloudExpressAsyncClient;
import org.jclouds.vcloud.terremark.TerremarkVCloudExpressClient;
import org.jclouds.vcloud.terremark.options.TerremarkInstantiateVAppTemplateOptions;

import com.google.common.base.Function;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;

public class VApp implements ServerServices {
    static private final Logger logger = Logger.getLogger(VApp.class);
    
    private Terremark provider = null;
    
    public VApp(Terremark t) {
        provider = t;
    }
    
    @Override
    public void boot(String serverId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            try {
                ctx.getApi().powerOnVApp(URI.create(serverId));
            }
            catch( RuntimeException e ) {
                logger.error(e);
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }
    }

    public Server clone(String serverId, String intoDcId, String name, String description, boolean powerOn, String ... firewallIds) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            try {
                CloneVAppOptions[] options = new CloneVAppOptions[powerOn ? 3 : 2];
                URI newServerId;
                Task task;
                    
                options[0] = CloneVAppOptions.Builder.withDescription(description);
                options[1] = CloneVAppOptions.Builder.deploy();
                if( powerOn ) {
                    options[2] = CloneVAppOptions.Builder.powerOn();
                }
                task = ctx.getApi().cloneVAppInVDC(URI.create(provider.getContext().getRegionId()),URI.create(serverId), name, options);
                waitFor(task);
                newServerId = task.getOwner().getHref();
                return getServer(newServerId.toASCIIString());
            }
            catch( RuntimeException e ) {
                logger.error(e);
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }        
    }
    
    @Override
    public Server define(String templateId, ServerSize size, String dataCenterId, String name, boolean withAnalytics, String ... firewalls) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        
        try {
            String[] parts = size.getSizeId().split(":");
            VCloudExpressVApp vapp;
            
            name = validateName(name, templateId);
            try {
                int memory = Integer.parseInt(parts[0]);
                int cpuCount = Integer.parseInt(parts[1]);
                String password = getRandomPassword();
                
                TerremarkInstantiateVAppTemplateOptions opt = TerremarkInstantiateVAppTemplateOptions.Builder.memory(memory).processorCount(cpuCount).withPassword(password);

                if( cpuCount < 1 ) {
                    cpuCount = 1;
                }
                vapp = ctx.getApi().instantiateVAppTemplateInVDC(URI.create(dataCenterId), URI.create(templateId),name,  opt);
                if( vapp == null ) {
                    return null;
                }
                URI serverId = vapp.getHref();
                
                waitFor(ctx.getApi().deployVApp(serverId));

                Server server = getServer(serverId.toASCIIString());

                server.setRootPassword(password);
                server.setImageId(templateId);
                return server;
            }
            catch( RuntimeException e ) {
                logger.error(e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public String getConsoleOutput(String serverId, long since) throws InternalException, CloudException {
        return "";
    }

    @Override
    public Collection<String> getFirewalls(String serverId) throws InternalException, CloudException {
        return new ArrayList<String>();
    }

    private HashMap<String,String> imageMap = new HashMap<String,String>();
    
    private String guessImageId(String osDescription) throws CloudException, InternalException {
        String searchString;

        if( osDescription == null ) {
            return null;
        }
        if( imageMap.containsKey(osDescription) ) {
            return imageMap.get(osDescription);
        }
        if( osDescription.contains("RHEL") || osDescription.contains("Red Hat") ) {
            searchString = "RHEL";
        }
        else if( osDescription.contains("CentOS") ) {
            searchString = "CentOS";
        }
        else if( osDescription.contains("Ubuntu") ) {
            searchString = "Ubuntu Server";
        }
        else if( osDescription.contains("Windows") && osDescription.contains("2003") ) {
            if( osDescription.contains("Enterprise") ) {
                searchString = "Windows 2003 Enterprise";
            }
            else {
                searchString = "Windows 2003 Standard";
            }
        }
        else if( osDescription.contains("Windows") && osDescription.contains("2008") ) {
            if( osDescription.contains("Enterprise") ) {
                searchString = "Windows Server 2008 Enterprise";
            }
            else if( osDescription.contains("Web") ) {
                searchString = "Windows Web Server 2008";
            }
            else {
                searchString = "Windows Server 2008 Standard";                
            }
        }
        else {
            searchString = osDescription;
        }
        boolean s32 = osDescription.contains("32-bit") || osDescription.contains("32 bit");
        for( ServerImage image : provider.getImageServices().list(provider.getContext().getAccountNumber()) ) {
            if( image.getDescription().startsWith(searchString) ) {
                if( s32 == image.getArchitecture().equals(Architecture.I32) ) {
                    imageMap.put(osDescription, image.getProviderServerImageId());
                    return image.getProviderServerImageId();
                }
            }
        }
        return null;
    }
    
    @Override
    public String getProviderTermForServer(Locale locale) {
        return "server";
    }

    static private final Random random = new Random();
    
    static public String alphabet = "ABCEFGHJKMNPRSUVWXYZabcdefghjkmnpqrstuvwxyz#@&()=+/";
    
    public String getRandomPassword() {
        StringBuilder password = new StringBuilder();
        int rnd = random.nextInt();
        int length = 8;
        
        if( rnd < 0 ) {
            rnd = -rnd;
        }
        length = length + (rnd%8);
        while( password.length() < length ) {
            char c;
            
            rnd = random.nextInt();
            if( rnd < 0 ) {
                rnd = -rnd;
            }
            c = (char)(rnd%255);
            if( alphabet.contains(String.valueOf(c)) ) {
                password.append(c);
            }
        }
        return password.toString();
    }
    
    @Override
    public Server getServer(String serverId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        
        try {
            VCloudExpressVApp server;
            
            try {
                server = ctx.getApi().getVApp(URI.create(serverId));
            }
            catch( RuntimeException e ) {
                String msg = e.getMessage();
                
                if( msg != null && msg.contains("NotFound") ) {
                    server = null;
                }
                else {
                    logger.error("Error loading server " + serverId + " from cloud: " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
            }
            if( server == null ) {
                return null;
            }
            return toServer(server);
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public ServerPersistence getServerPersistence() throws InternalException, CloudException {
        return ServerPersistence.ALL;
    }

    @Override
    public ServerStatistics getServerStatistics(String serverId, long start, long end) throws InternalException, CloudException {
        return new ServerStatistics();
    }

    @Override
    public Collection<ServerStatistics> getServerStatisticsForPeriod(String serverId, long start, long end) throws InternalException, CloudException {
        return new ArrayList<ServerStatistics>();
    }

    private ServerSize getSize(VCloudExpressVApp server2) {
        Multimap<ResourceType,ResourceAllocation> resourceAllocation = Multimaps.index(server2.getResourceAllocations(), new Function<ResourceAllocation, ResourceType>(){

         @Override
         public ResourceType apply(ResourceAllocation arg0) {
            return arg0.getType();
         }
           
        });
        ResourceAllocation memory, processor;
        String str;
        
        if( resourceAllocation == null ) {
            return null;
        }
        Collection<ResourceAllocation> alloc = resourceAllocation.get(ResourceType.MEMORY);
        
        if( alloc != null && alloc.size() > 0 ) {
            memory = alloc.iterator().next();
        }
        else {
            memory = null;
        }
        alloc = resourceAllocation.get(ResourceType.PROCESSOR);
        if( alloc != null && alloc.size() > 0 ) {
            processor = alloc.iterator().next();
        }
        else {
            processor = null;
        }
        if( memory == null || processor == null ) {
            return null;
        }
        str = memory.getVirtualQuantity() + ":" + processor.getVirtualQuantity();
        return new ServerSize(str, processor.getVirtualQuantity() + " CPU, " + memory.getVirtualQuantity() + "MB RAM ");
    }
    
    @Override
    public Collection<ServerSize> getSupportedSizes(Architecture architecture) throws InternalException, CloudException {
        ArrayList<ServerSize> sizes = new ArrayList<ServerSize>();
        
        sizes.add(new ServerSize("512:1", "1 CPU, 512M RAM, 4GB Disk"));
        sizes.add(new ServerSize("512:2", "2 CPU, 512M RAM, 4GB Disk"));
        sizes.add(new ServerSize("512:4", "4 CPU, 512M RAM, 4GB Disk"));
        sizes.add(new ServerSize("512:8", "8 CPU, 512M RAM, 4GB Disk"));

        sizes.add(new ServerSize("1024:1", "1 CPU, 1GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("1024:2", "2 CPU, 1GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("1024:4", "4 CPU, 1GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("1024:8", "8 CPU, 1GB RAM, 4GB Disk"));

        sizes.add(new ServerSize("1536:1", "1 CPU, 1.5GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("1536:2", "2 CPU, 1.5GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("1536:4", "4 CPU, 1.5GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("1536:8", "8 CPU, 1.5GB RAM, 4GB Disk"));

        sizes.add(new ServerSize("2048:1", "1 CPU, 2GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("2048:2", "2 CPU, 2GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("2048:4", "4 CPU, 2GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("2048:8", "8 CPU, 2GB RAM, 4GB Disk"));

        sizes.add(new ServerSize("4096:1", "1 CPU, 4GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("4096:2", "2 CPU, 4GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("4096:4", "4 CPU, 4GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("4096:8", "8 CPU, 4GB RAM, 4GB Disk"));

        sizes.add(new ServerSize("8192:1", "1 CPU, 8GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("8192:2", "2 CPU, 8GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("8192:4", "4 CPU, 8GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("8192:8", "8 CPU, 8GB RAM, 4GB Disk"));

        sizes.add(new ServerSize("12288:1", "1 CPU, 12GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("12288:2", "2 CPU, 12GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("12288:4", "4 CPU, 12GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("12288:8", "8 CPU, 12GB RAM, 4GB Disk"));

        sizes.add(new ServerSize("16384:1", "1 CPU, 16GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("16384:2", "2 CPU, 16GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("16384:4", "4 CPU, 16GB RAM, 4GB Disk"));
        sizes.add(new ServerSize("16384:8", "8 CPU, 16GB RAM, 4GB Disk"));
        
        return sizes;
    }
    
    @Override
    public boolean isSupportsCloning() {
        return true;
    }

    @Override
    public Server launch(String templateId, ServerSize size, String dataCenterId, String serverName, String withKey, String inVlanId, boolean withMonitoring, String... firewalls) throws InternalException, CloudException {
        Server server = define(templateId, size, dataCenterId, serverName, withMonitoring, firewalls);
        
        boot(server.getProviderServerId());
        return server;
    }

    @Override
    public Collection<Server> list() throws InternalException, CloudException {
        ArrayList<Server> servers = new ArrayList<Server>();
        Vdc dsnVdc = (Vdc)provider.getDataCenterServices();

        for( Region region : dsnVdc.listRegions() ) {
            for( DataCenter dc : dsnVdc.listDataCenters(region.getProviderRegionId()) ) {
                Collection<ReferenceType> resources;
            
                try {
                    VDC vdc = dsnVdc.getVdc(dc.getProviderDataCenterId());
                    
                    if (vdc.getResourceEntities() != null) {
                        resources = vdc.getResourceEntities().values();
                    } else {
                        resources = Collections.emptyList();
                    }
                }
                catch( RuntimeException e ) {
                    logger.error("Could not load servers: " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
                for( ReferenceType resource : resources ) { 
                    if( resource.getType().equals(VCloudMediaType.VAPP_XML) ) {
                        Server server = getServer(resource.getHref().toASCIIString());
                        
                        if( server != null ) {
                            servers.add(server);
                       }
                    }
                }
            }
        }
        return servers;
    }

    @Override
    public void monitor(String serverId) throws InternalException, CloudException {
        // NO-OP
    }

    @Override
    public void pause(String serverId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            try {
                ctx.getApi().powerOffVApp(URI.create(serverId));
            }
            catch( RuntimeException e ) {
                logger.error(e);
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void reboot(String serverId) throws CloudException, InternalException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            ctx.getApi().resetVApp(URI.create(serverId));
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void stop(String serverId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            waitFor(ctx.getApi().powerOffVApp(URI.create(serverId)));
            try {
                ((PublicIp)provider.getAddressServices()).clear(serverId);
                ctx.getApi().deleteVApp(URI.create(serverId));
            }
            catch( RuntimeException e ) {
                logger.error(e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void unmonitor(String serverId) throws InternalException, CloudException {
        // NO-OP
    }

    private Server toServer(VCloudExpressVApp server2) throws CloudException, InternalException {
        Server server = new Server();
        ServerState currentState;
        
        server.setProviderServerId(server2.getHref().toASCIIString());
        switch ( server2.getStatus() ) {
            case ON: currentState = ServerState.RUNNING; break;
            case OFF: currentState = ServerState.PAUSED; break;
            case SUSPENDED: currentState = ServerState.PAUSED; break;
            default: currentState = ServerState.PENDING; break;
        }
        server.setCurrentState(currentState);        
        server.setName(server2.getName());
        String str = server2.getOperatingSystemDescription();
        if( str == null || str.indexOf("32-bit") != -1 || str.indexOf("32 bit") != -1 ) {
            server.setArchitecture(Architecture.I32);            
        }
        else if( str.indexOf("64-bit") != -1 || str.indexOf("64 bit") != -1 ) {
            server.setArchitecture(Architecture.I64);
        }
        else {
            server.setArchitecture(Architecture.I64);
        }
        server.setCreateTime(0L);
        server.setDescription(server2.getName() + " - " + server2.getOperatingSystemDescription());
        server.setPersistent(true);
        server.setPlatform(Platform.guess(server2.getOperatingSystemDescription()));
        ServerSize size = getSize(server2);
        
        if( size == null ) {
            return null;
        }
        else {
            server.setSize(size.getSizeId());
        }
        String name = server2.getName();
        int idx = name.lastIndexOf("-");
        
        String templateId = null;
        
        if( idx != -1 && idx < (name.length()-1) ) {
            String tid = name.substring(idx+1);
            
            try {
                Integer.parseInt(tid);
                templateId = tid;
            }
            catch( NumberFormatException ignore ) {
                // ignore me
            }
        }
        if( templateId != null ) {
            server.setImageId(templateId);
        }
        else {
            server.setImageId(guessImageId(server2.getOperatingSystemDescription()));
        }
        Collection<String> addresses;
        
        try {
            addresses = server2.getNetworkToAddresses().values();
        }
        catch( RuntimeException e ) {
            logger.error(e.getMessage());
            e.printStackTrace();
            throw new CloudException(e);
        }
        if( addresses == null || addresses.size() < 1 ) {
            server.setPrivateIpAddresses(new String[0]);
        }
        else {
            String[] ips = new String[addresses.size()];
            int i = 0;
            
            for( String addr : addresses ) {
                if( i == 0 ) {
                    server.setPrivateDnsAddress(addr);
                }
                ips[i++] = addr;
            }
            server.setPrivateIpAddresses(ips);
        }
        server.setPublicDnsAddress(null);
        server.setPublicIpAddresses(new String[0]);
        server.setRegionId("us-miami");
        server.setDataCenterId(server2.getVDC().getHref().toASCIIString());
        server.setReservedAddressId(null);
        server.setRootPassword(((Template)provider.getImageServices()).getPassword(server.getImageId()));
        server.setStopTime(-1L);
        server.setTerminationTime(-1L);
        server.setStartTime(0L);
        return server;
    }
    
    private String validateName(String name, String suffix) {
        int len = (suffix == null ? 0 : (1+suffix.length()));
        
        name = name.toLowerCase().replaceAll("_", "-");
        if( name.length() <= (15-len) ) {
            return (suffix == null ? name : name + "-" + suffix);
        }
        return name.substring(0, 15-(len+1)) + (suffix == null ? "" : "-" + suffix);
    }
    
    private void waitFor(Task task) throws CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            TaskStatus status = task.getStatus();
            
            while( status.equals(TaskStatus.RUNNING) || status.equals(TaskStatus.QUEUED) ) {
                if( status.equals(TaskStatus.ERROR) || status.equals(TaskStatus.UNRECOGNIZED) ) {
                    throw new CloudException(task.getOwner().getName());
                }
                try { Thread.sleep(1000L); }
                catch( InterruptedException e ) { }
                task = ctx.getApi().getTask(task.getHref());
                status = task.getStatus();
            }        
        }
        finally {
            ctx.close();
        }
    }
}
